package provide OSM 0.1

package require http
package require tdom
package require TclOO

package require OSM::canvas

if {[catch {package require Dict}]} {
    namespace eval ::tcl::dict {
           # dict get? courtesy patthoyts
            proc get? {dict args} {
            if {[dict exists $dict {*}$args]} {
                return [dict get $dict {*}$args]
            } else {
                return {}
            }
        }
        foreach x {get?} {
            namespace ensemble configure dict -map [linsert [namespace ensemble configure dict -map] end $x ::tcl::dict::$x]
        }
        ::unset x
    }
}

oo::class create OSM {
    variable storage styles

    constructor {args} {
        foreach {k v} $args {
            set [string trimleft $k -] $v
        }

        if {[dict exists $args canvas] || [dict exists $args -canvas]} {
            set storage [OSM::canvas new {*}$args owner [namespace which my]]
        } else {
            set storage [OSM::DOM new {*}args]
        }
    }

    method node {args} {
        set args [lassign $args id op]
        {*}[concat $storage node-$op $id $args]
    }

    method way {args} {
        set args [lassign $args id op]
        {*}[concat $storage way-$op $id $args]
    }

    method relation {args} {
        set args [lassign $args id op]
        {*}[concat $storage relation-$op $id $args]
    }

    method parse {args} {
        $storage parse {*}$args
    }

    method map-to-screen {args} {
        $storage map-to-screen {*}$args
    }

    # style items:
    # 1. what tags to glob-match
    # 2. what tags not to glob-match
    # 3. what style to apply: list of (object type plus options (in a single list))

    # needs to be rewritten to support zoom-dependent width;
    # also, to support MapCSS, some features are needed

    # set styles {
    #     {highway *} {} {line -smooth no -width 3}
    #     {railway *} {} {line -smooth no -width 3 -dash -}
    #     {building *} {} {polygon -smooth no -fill brown}
    #     {} {area yes} polygon
    #     {landuse forest} {} {polygon -fill darkgreen}
    #     {natural water} {} {polygon -fill darkblue}
    #     {waterway *} {} {line -fill darkblue -smooth yes}
    #     {waterway riverbank} {} polygon
    # }

    method getstyle {type tags} {
        set options {}
        set object {}
        set itemtags {}
        foreach {pospat negpat style} $styles {
            set posmatch 0
            set negmatch 1
            set soptions [lassign $style sobject]
            set sitemtags [lassign [split $sobject @] sobject]

            foreach {poskey posval} $pospat {
                # absent positive match key is treated as fail
                if {[string match %* $poskey] && ($object eq [string trimleft $poskey %])} {
                    incr posmatch
                } elseif {[string match -* $poskey] && ($type eq [string trimleft $poskey -])} {
                    incr posmatch
                } elseif {$poskey eq "*"} {
                    if {[llength $tags] != 0} {
                        incr posmatch
                    } else {
                        set posmatch 0
                    }
                } elseif {[dict exists $tags $poskey]} {
                    if {[string match $posval [dict get $tags $poskey]]} {
                        incr posmatch
                    } else {
                        set posmatch 0
                    }
                } else {
                    set posmatch 0
                }
            }

            foreach {negkey negval} $negpat {
                # absent negative match key is treated as match
                if {[string match %* $negkey] && ($object eq [string trimleft $negkey %])} {
                    set negmatch 0
                } elseif {[string match -* $negkey] && ($type eq [string trimleft $negkey -])} {
                    set negmatch 0
                } elseif {$negkey eq "*"} {
                    if {[llength $tags] == 0} {
                        incr negmatch
                    } else {
                        set negmatch 0
                    }
                } elseif {[dict exists $tags $negkey]} {
                    if {[string match $negval [dict get $tags $negkey]]} {
                        set negmatch 0
                    } else {
                        incr negmatch
                    }
                } else {
                    incr negmatch
                }
            }

            if {$posmatch && $negmatch} {
                # here we need at least three options:
                #  +* adds some options but doesn't change the object type
                #  -* overrides any previous style settings
                #  anything else overrides the object, adds some options
                switch -glob $sobject {
                    {+*} {
                        lappend options {*}$soptions
                        lappend itemtags {*}$sitemtags
                    }
                    {-*} {
                        set options $soptions
                        set object $sobject
                        set itemtags $sitemtags
                    }
                    default {
                        lappend options {*}$soptions
                        lappend itemtags {*}$sitemtags
                        set object $sobject
                    }
                }
            }
        }

        if {$itemtags ne ""} {
            set itemtags @@[string map {{ } { @@}} $itemtags]
        }

        return [list [list $object $options $itemtags]]
    }
}

oo::class create OSM::DOM {
    variable dom doc diffmode

    constructor {args} {
        foreach {k v} $args {
            set [string trimleft - $k] $v
        }
        set diffmode 1
    }

    method select {args} {
        # need full mapcss selector support here
        my variable doc
        lassign $args type
        switch $type {
            node -
            way {
                return [string map {{id } {} \{ {} \} {}} [$doc selectNodes "/osm/$type/@id"]]
            }
        }
    }

    method foreach {selector args} {
        set parenttype {}
        set parentid {}
        set lambda {}
        set compact 0
        foreach arg $args {
            switch $arg {
                in          {}
                way         -
                relation    {
                    set parenttype $arg
                }
                compact     {
                    set compact 1
                }
                default     {
                    if {($parentid ne "") || ($parenttype eq "")} {
                        set lambda $arg
                    } else {
                        set parentid $arg
                    }
                }
            }
        }
        my variable doc
        if {$parentid ne ""} {
            set membertype [string map {way nd relation member} $parenttype]
            # the following line is wrong when selector in not just way or node or relation
            set submembertype [string map {way nodes relation members} $selector]
            foreach member [$doc selectNodes "/osm/$parenttype\[@id=$parentid\]/$membertype"] {
                set member [$doc selectNodes "/osm/$selector\[@id=[$member @ref]\]"]
                #
                if {$compact} {
                    lassign [$member asList] -> member
                } else {
                    lassign [$member asList] -> member objs
                    set members {}
                    foreach obj $objs {
                        lassign $obj tag attrs
                        if {$tag eq "tag"} {
                            dict set member tags [dict get $attrs k] [dict get $attrs v]
                        } else {
                            lappend members [dict get $attrs ref]
                        }
                    }
                    dict set member $submembertype $members
                }
                eval $lambda
            }
        }
    }

    method parse {args} {
        my variable doc
        set dom [dom parse {*}$args]
        set doc [$dom documentElement]
    }

    # export all data
    #   -changesonly 1 to export only objects were modified
    method export {args} {
    }

    # create a new node
    # new lat 0 lon 0 ... tags {k1 v1 k2 v2...}
    method node-new {args} {
        # set node [$doc selectNodes {/osm/node[@id=$id]}]
    }

    method node-delete {id} {
        my variable doc
        set node [$doc selectNodes {/osm/node[@id=$id]}]
        if {$node eq ""} {
            error "wrong node id"
        } else {
            $node delete
            $doc appendFromList "node {id $id action delete} {}"
            return $id
        }
    }

    method node-update {id} {
        my variable doc
        set node [$doc selectNodes {/osm/node[@id=$id]}]
        if {$node eq ""} {
            # maybe throw error here?
            $doc appendFromList "node {id $id action modify} {}"
        } else {
            $node setAttribute action modify
            return $id
        }
    }

    method node-dom-get {domid} {
        lassign [$domid asList] -> node tags
        foreach tag $tags {
            lassign $tag -> tag
            dict set node tags [dict get $tag k] [dict get $tag v]
        }
        return $node
    }

    # get the node or its parameter:
    #   node get $id: returns a list containing all info about the node
    #   node get $id lat: returns lat attribute
    #   node get $id tags highway: returns value of a tag named highway
    method node-get {id args} {
        my variable doc
        switch [llength $args] {
            2 {
                lassign $args -> name
                set tag [$doc selectNodes {/osm/node[@id=$id]/tag[@k=$name]}]
                if {$tag eq ""} {
                    return ""
                } else {
                    return [$tag @v ""]
                }
            }
            1 {
                lassign $args attr
                if {$attr ne "tags"} {
                    set node [$doc selectNodes {/osm/node[@id=$id]}]
                    if {$node eq ""} {
                        return ""
                    } else {
                        set result {}
                        foreach a $attr {
                            lappend result [$node @$a]
                        }
                        return $result
                    }
                } else {
                    set tags {}
                    foreach tag [$doc selectNodes {/osm/node[@id=$id]/tag}] {
                        lappend tags [$tag @k] [$tag @v]
                    }
                    return $tags
                }
            }
            0 {
                set node [$doc selectNodes {/osm/node[@id=$id]}]
                if {$node eq ""} {
                    return ""
                } else {
                    lassign [$node asList] -> node tags
                    foreach tag $tags {
                        lassign $tag -> tag
                        dict set node tags [dict get $tag k] [dict get $tag v]
                    }
                    return $node
                }
            }
            default {
                error "wrong number of args"
            }
        }
    }

    method node-set {id args} {
        my variable doc
        switch [llength $args] {
            3 {
                lassign $args -> name value
                set node [$doc selectNodes {/osm/node[@id=$id]}]
                if {$node eq ""} {
                    error "wrong node id"
                } else {
                    node-update $id
                    set tagnode [$node find k $name]
                    if {$tagnode ne ""} {
                        if {$value ne ""} {
                            $tagnode setAttribute v $value
                        } else {
                            $tagnode delete
                        }
                        return $value                        
                    } else {
                        $node appendFromList [string map [list %NAME $name %VALUE $value] {tag {k %NAME v %VALUE} {}}]
                    }
                }
            }
            2 {
                lassign $args attr value
                set node [$doc selectNodes [string map [list %ID $id] {/osm/node[@id="%ID"]}]]
                if {$node eq ""} {
                    error "wrong node id"
                } else {
                    node-update $id
                    $node setAttribute $attr $value
                    return $value
                }
            }
            1 {
                set node [$doc selectNodes {/osm/node[@id=$id]}]
                if {$node eq ""} {
                    error "wrong node id"
                } else {
                    $node delete
                    lassign $args args
                    $doc appendFromList [list node [dict remove $args tags] {}]
                    node-update $id
                    set node [$doc selectNodes {/osm/node[@id=$id]}]
                    foreach tag [dict get? $args tags] {
                        $node appendFromList [string map [list %NAME [dict get $tag k] %VALUE [dict get $tag v]] {tag {k %NAME v %VALUE} {}}]
                    }
                    return $args
                }
            }
            default {
                error "wrong number of args"
            }
        }
    }

    method way-update {id} {
        my variable doc
        set way [$doc selectNodes {/osm/way[@id=$id]}]
        if {$way eq ""} {
            # maybe throw error here?
            $doc appendFromList "way {id $id action modify} {}"
        } else {
            $way setAttribute action modify
            return $id
        }
    }

    # get the node or its parameter:
    #   way get $id: returns a list containing all info about the node
    #   way get $id lat: returns lat attribute
    #               ^- this can be a list
    #   way get $id tags highway: returns value of a tag named highway
    #   way get $id nodes: returns the list of nodes
    #   way get $id nodes 0: returns the first node in the way
    method way-get {id args} {
        my variable doc
        switch [llength $args] {
            2 {
                lassign $args attr name
                switch $attr {
                    tag     -
                    tags    {
                        set way [$doc selectNodes {/osm/way[@id=$id]}]
                        set tag [$way child all tag k $name]
                        if {$tag eq ""} {
                            return ""
                        } else {
                            return [$tag @v ""]
                        }
                    }
                    node    -
                    nodes   {
                        set way [$doc selectNodes {/osm/way[@id=$id]}]
                        set nd [$way child [::tcl::mathop::+ 1 $name] nd]
                        if {$nd eq ""} {
                            return ""
                        } else {
                            return [$nd @ref ""]
                        }
                    }
                }
            }
            1 {
                lassign $args attr
                switch $attr {
                    tag     -
                    tags    {
                        set tags {}
                        set way [$doc selectNodes {/osm/way[@id=$id]}]
                        foreach tag [$way child all tag] {
                            lappend tags [$tag @k] [$tag @v]
                        }
                        return $tags
                    }
                    node    -
                    nodes   {
                        set nodes {}
                        set way [$doc selectNodes {/osm/way[@id=$id]}]
                        foreach node [$way child all nd] {
                            lappend nodes [$node @ref]
                        }
                        return $nodes
                    }
                    default {
                        set way [$doc selectNodes {/osm/way[@id=$id]}]
                        if {$way eq ""} {
                            return ""
                        } else {
                            set result {}
                            foreach a $attr {
                                lappend result [$way @$a]
                            }
                            return $result
                        }
                    }
                }
            }
            0 {
                set way [$doc selectNodes {/osm/way[@id=$id]}]
                if {$way eq ""} {
                    return ""
                } else {
                    lassign [$way asList] -> way objs
                    foreach obj $objs {
                        lassign $obj tag attrs
                        if {$tag eq "tag"} {
                            dict set way tags [dict get $attrs k] [dict get $attrs v]
                        } else {
                            lappend nodes [dict get $attrs ref]
                        }
                    }
                    dict set way nodes $nodes
                    return $way
                }
            }
            default {
                error "wrong number of args"
            }
        }
    }

    method way-set {id args} {
        my variable doc
        switch [llength $args] {
            3 {
                # no nodes update yet
                lassign $args attr name value
                set way [$doc selectNodes {/osm/way[@id=$id]}]
                if {$way eq ""} {
                    error "wrong way id"
                } else {
                    way-update $id
                    set tagnode [$way find k $name]
                    if {$tagnode ne ""} {
                        if {$value ne ""} {
                            $tagnode setAttribute v $value
                        } else {
                            $tagnode delete
                        }
                        return $value                        
                    } else {
                        $way appendFromList [string map [list %NAME $name %VALUE $value] {tag {k %NAME v %VALUE} {}}]
                    }
                }
            }
            2 {
                lassign $args attr value
                set node [$doc selectNodes [string map [list %ID $id] {/osm/way[@id="%ID"]}]]
                if {$node eq ""} {
                    error "wrong way id"
                } else {
                    node-update $id
                    $node setAttribute $attr $value
                    return $value
                }
            }
            1 {
                set node [$doc selectNodes {/osm/way[@id=$id]}]
                if {$node eq ""} {
                    error "wrong way id"
                } else {
                    $node delete
                    lassign $args args
                    $doc appendFromList [list way [dict remove $args tags] {}]
                    node-update $id
                    set node [$doc selectNodes {/osm/way[@id=$id]}]
                    foreach tag [dict get? $args tags] {
                        $node appendFromList [string map [list %NAME [dict get $tag k] %VALUE [dict get $tag v]] {tag {k %NAME v %VALUE} {}}]
                    }
                    return $args
                }
            }
            default {
                error "wrong number of args"
            }
        }
    }

}

if {$tcl_interactive} {
#    set osm [::OSM new]
#    $osm parse {<osm><node id="23"><tag k="highway" v="cycleway"/><tag k="bicycle" v="yes"/></node><node id="42"/><way id="32"><tag k="r" v="e"/><nd ref="23"/><tag k="h" v="c"/></way></osm>}
}

# vim: set ts=4:sw=4:sts=4
