package provide OSM::canvas 0.1

package require OSM::Icons

oo::class create OSM::canvas {
    variable canvas diffmode nodes waynodes relations current currentobject owner

    constructor {args} {
        set owner my
        set diffmode 1
        foreach {k v} $args {
            set [string trimleft $k -] $v
        }
    }

    method parse {args} {
        set parser [xml::parser]
        $parser configure -elementstartcommand [list [namespace which my] hs] -elementendcommand [list [namespace which my] he]
        if {(([llength $args] % 2)==0) && [dict exists $args -channel]} {
            $parser parsechannel [dict get $args -channel]
        } else {
            $parser parse {*}$args
        }
    }

    method node-get {id args} {
        if {![info exists nodes($id)]} {
            return ""
        }
        switch [llength $args] {
            2 -
            1 {
                dict get? $nodes($id) {*}$args
            }
            0 {
                return $nodes($id)
            }
            default {
                error "wrong number of args"
            }
        }
    }

    method node-set {id args} {
        if {![info exists nodes($id)]} {
            return ""
        }
        switch [llength $args] {
            3 -
            2 {
                dict set nodes($id) {*}$args
            }
            1 {
                set nodes($id) {*}$args
            }
            default {
                error "wrong number of args"
            }
        }
    }

    method way-get {id args} {
        set waylist [$canvas find withtag @id=$id&&@@ways]
        if {![llength $waylist]} {
            return ""
        }
        lassign $waylist wayitem
        switch [llength $args] {
            2 {
                lassign $args attrs name
                switch $attrs {
                    tags {
                        ### TODO: fix the split
                        return [lassign [split [lsearch -inline -glob [$canvas gettags $wayitem] @$name=*] #] -]
                    }
                    nodes {
                        return [lindex $waynodes($id) $name]
                    }
                    default {
                        return ""
                    }
                }
            }
            1 {
                lassign $args attr
                switch $attr {
                    tags {
                        return [dict filter [my canvas-to-dict [$canvas gettags $wayitem]] key {[A-Z_a-z0-9]*}]
                    }
                    nodes {
                        return $waynodes($id)
                    }
                    default {
                        return [dict get? [my canvas-to-dict [$canvas gettags $wayitem]] @$attr]
                    }
                }
            }
            0 {
                foreach {k v} [my canvas-to-dict [$canvas gettags $wayitem]] {
                    if {[string match @* $k]} {
                        dict set result [string trimleft $k @] $v
                    } else {
                        dict set result tags $k $v
                    }
                }
                return $result
            }
        }
    }

    method way-set {id args} {
        set waylist [$canvas find withtag @id=$id&&@@ways]
        if {![llength $waylist]} {
            return ""
        }
        lassign $waylist wayitem
        switch [llength $args] {
            3 {
                lassign $args attrs k v
                switch $attrs {
                    tags {
                        $canvas dtag $wayitem [lsearch -inline -glob [$canvas gettags $wayitem] $k=*]
                        if {$v ne ""} {
                            $canvas addtag $k=$v withtag $wayitem
                        }
                    }
                    nodes {
                        lset waynodes($id) $k $v
                        # TODO: update the canvas!
                    }
                    default {
                        return ""
                    }
                }
            }
            2 {
                lassign $args attr value
                switch $attr {
                    tags {
                        set itemtags [lsearch -inline -all -glob [$canvas gettags $wayitem @*]]
                        foreach {k v} $value {
                            lappend itemtags $k=$v
                        }
                        $canvas itemconfigure $wayitem -tags $itemtags
                    }
                    nodes {
                        set waynodes($id) $value
                    }
                    default {
                        $canvas dtag $wayitem [lsearch -inline -glob [$canvas gettags $wayitem] @$k=*]
                        if {$v ne ""} {
                            $canvas addtag @$k=$v withtag $wayitem
                        }
                    }
                }
            }
            1 {
                lassign $args value
                set waynodes($id) [dict get? $value nodes]
                # TODO: update the canvas!
                set itemtags [lsearch -inline -all -glob [$canvas gettags $wayitem @@*]]
                foreach {k v} [dict get? $value tags] {
                    lappend itemtags $k=$v
                }
                dict unset value tags
                dict unset value nodes
                foreach {k v} $value {
                    lappend itemtags @$k=$v
                }
                $canvas itemconfigure $wayitem -tags $itemtags
                # TODO: update the canvas using [getstyle]!
            }
        }
    }


    method canvas-to-dict {taglist} {
        foreach tag $taglist {
            ### TODO: this can and will break on values containing equal signs
            lassign [split $tag =] key value
            if {($key eq "") || ($value eq "")} continue
            dict set result $key $value
        }
        return $result
    }

    method relation-get {id args} {
        if {![info exists relations($id)]} {
            return ""
        }
        switch [llength $args] {
            2 -
            1 {
                return [dict get? $relations($id) {*}$args
            }
            0 {
                return $relations($id)
            }
            default {
                error "wrong number of args"
            }
        }

    }

    method hs {name attlist} {
        lappend current $name
        switch $name {
            node     -
            way      -
            relation {
                set currentobject $attlist
            }
            tag     {
                dict set currentobject tags [dict get $attlist k] [dict get $attlist v]
            }
            nd      {
                dict lappend currentobject nodes [dict get $attlist ref]
            }
            member  {
                dict lappend currentobject members $attlist
            }
        }
    }

    method he {name} {
        set current [lreplace $current end end]
        switch $name {
            node     {
                set [set name]s([dict get $currentobject id]) $currentobject
                if {yes} {
                    set lat [dict get $currentobject lat]
                    set lon [dict get $currentobject lon]
                    lassign [$owner map-to-screen $lat $lon] y x
                    set styles [$owner getstyle node [dict get? $currentobject tags]]
                    lassign $styles style ; # leave it so for now
                    lassign $style object options tags
                    if {$object ne ""} {
                        set textoptions [dict get? $options -textoptions]
                        set state [if {[dict exists $options -state]} {
                            dict get? $options -textoptions
                        } else {
                            return -level 0 normal
                        }]
                        set tagadd [expr {($state ne "hidden")?"@@hidden":""}]
                        dict unset options -textoptions
                        $canvas create $object $x $y -tags [list @@data @@nodes @id=[dict get $currentobject id] {*}$tags {*}$tagadd] {*}$options
                        set text [dict get? $currentobject tags name]
                        if {($textoptions ne "") && ($text ne "")} {
                            $canvas create text $x $y -tags [list @@data @@nodetexts @id=[dict get $currentobject id] {*}$tagadd] -text $text {*}$textoptions
                        }
                    }
                }
            }
            relation {
                set [set name]s([dict get $currentobject id]) $currentobject
            }
            way     {
                foreach id [dict get? $currentobject nodes] {
                    set lat [dict get $nodes($id) lat]
                    set lon [dict get $nodes($id) lon]
                    lassign [$owner map-to-screen $lat $lon] y x
                    lappend coords $x $y
                    $canvas addtag @@way=[dict get $currentobject id] withtag @@nodes&&@id=$id
                }
                set waynodes([dict get $currentobject id]) [dict get? $currentobject nodes]
                set tags [dict get? $currentobject tags]
                dict unset currentobject tags
                dict unset currentobject nodes
                foreach {k v} $currentobject {
                    lappend itemtags @$k=$v
                }
                foreach {k v} $tags {
                    lappend itemtags $k $k=$v
                }
                lappend itemtags @@data @@ways
                set styles [$owner getstyle way $tags]
                foreach style $styles {
                    lassign $style object options tags
                    if {$object ne ""} {
                        set objecttags $itemtags
                        if {$tags ne ""} {
                            lappend objecttags {*}$tags
                        }
                        $canvas create $object $coords {*}$options -tags $objecttags
                    }
                }
            }
        }
    }

    method map-to-screen {lat lon} {
        return [lrange [::map::slippy geo 2point [list 18 $lat $lon]] 1 2]
    }

    # returns [list object1 object2 ...]
    # each objectn is [list object-type options tags]
    method getstyle {type tags} {
        return {{line {}}}
    }
}


