package provide OSM::Editor::Presets 0.1

namespace eval OSM::Editor::Presets::Tags {
    variable Values {}
    variable Reverse_Map {}

    proc add {tags {reverse_map {}}} {
        variable Values
        variable Reverse_Map
        ::set Values [dict merge $Values $tags]
        ::set Reverse_Map [dict merge $Reverse_Map $reverse_map]
    }

    proc get args {
        variable Values
        dict get $Values {*}$args
    }

    proc set args {
        variable Values
        dict set Values {*}$args
    }

    proc get-reverse-map args {
        variable Reverse_Map
        dict get $Reverse_Map {*}$args
    }

    proc set-reverse-map args {
        variable Reverse_Map
        dict set Reverse_Map {*}$args
    }
}

package require OSM::Editor::Presets::BuiltIn

oo::class create OSM::Editor::Presets {
    variable canvas pane tagk tagv tags path detailsshown
    constructor {args} {
        global gui modulesd
        foreach {k v} $args {
            set [string trimleft $k -] $v
        }

        # 
        set pane [dict get $gui pane]
        ttk::notebook $pane.presets
        ttk::frame $pane.presets.add

        $pane.presets add $pane.presets.add -text {Add tags}
        # -padding {0 0 0 64}
        ttk::combobox $pane.presets.add.cb0 -state readonly -values [lsearch -not -glob -inline -all [dict keys [OSM::Editor::Presets::Tags::get]] -*]
        bind $pane.presets.add.cb0 <<ComboboxSelected>> [list [namespace which my] comboselected %W]
        set path {}
        set detailsshown no

        pack $pane.presets -fill x

        ttk::frame $pane.presets.add.details -padding 5
        pack $pane.presets.add.details -fill x
        ttk::button $pane.presets.add.apply -text Apply
        pack $pane.presets.add.apply

        [dict get $modulesd EditWay] configure -selectcommand [list [namespace which my] updatetags]
    }

    method addfield {m i} {
        if 1 {
            ttk::label $pane.presets.add.details.l$i -text [dict get? $m -name]
            if {[dict exists $m -suggest]} {
                ttk::combobox $pane.presets.add.details.e$i
                $pane.presets.add.details.e$i set [dict get? $m -default]
                $pane.presets.add.details.e$i configure -values [dict get? $m -suggest]
            } elseif {[dict get? $m -boolean] eq "yes"} {
                $pane.presets.add.details.l$i configure -text {}
                ttk::checkbutton $pane.presets.add.details.e$i -text [dict get? $m -name] -onvalue yes -offvalue no
                set ::$pane.presets.add.details.e$i [dict get? $m -default]
            } else {
                ttk::entry $pane.presets.add.details.e$i
                $pane.presets.add.details.e$i insert 0 [dict get? $m -default]
            }
        }
        grid $pane.presets.add.details.l$i -sticky w -column 0 -padx 4 -row $i
        grid $pane.presets.add.details.e$i -sticky we -column 1 -padx 4 -row $i
        grid columnconfigure $pane.presets.add.details $pane.presets.add.details.e$i -weight 1        
    }

    method makecombo path {
        set level [llength $path]
        set preset [OSM::Editor::Presets::Tags::get {*}$path]
        set values [lsearch -not -glob -inline -all [dict keys $preset] -*]
        if {$detailsshown} {
            set wl [info commands $pane.presets.add.details.l*]
            lappend wl {*}[info commands $pane.presets.add.details.e*]
            foreach w $wl {
                rename $w {}
            }
            set detailsshown no
        }
        if {![llength $values]} {
            set options [dict filter $preset key -*]
            foreach k {-type -title -tags -mandatory -optional} {
                set [string trimleft $k -] [dict get? $options $k]
            }
            ttk::label $pane.presets.add.details.lo
            if {[llength $optional]} {
                $pane.presets.add.details.lo configure -text Optional
            }
            set i 0
            foreach m $mandatory {
                my addfield $m $i
                grid $pane.presets.add.details.l$i -sticky w -column 0 -padx 4 -row $i
                grid $pane.presets.add.details.e$i -sticky we -column 1 -padx 4 -row $i
                grid columnconfigure $pane.presets.add.details $pane.presets.add.details.e$i -weight 1
                incr i
            }
            grid $pane.presets.add.details.lo -
            incr i
            foreach m $optional {
                my addfield $m $i
                grid $pane.presets.add.details.l$i -sticky w -column 0 -padx 4 -row $i
                grid $pane.presets.add.details.e$i -sticky we -column 1 -padx 4 -row $i
                grid columnconfigure $pane.presets.add.details $pane.presets.add.details.e$i -weight 1
                incr i
            }
            set detailsshown yes
            return
        }
        if {![llength [info commands $pane.presets.add.cb$level]]} {
            ttk::combobox $pane.presets.add.cb$level -state readonly -values $values
            bind $pane.presets.add.cb$level <<ComboboxSelected>> [list [namespace which my] comboselected %W]
            pack $pane.presets.add.cb$level -fill x -before $pane.presets.add.details
        } else {
            $pane.presets.add.cb$level configure -state readonly -values $values
            $pane.presets.add.cb$level set {}
            pack $pane.presets.add.cb$level -fill x -before $pane.presets.add.details

        }
    }

    method comboselected W {
        set no [string index $W end]
        set l [llength $path]
        if {$l <= $no} {
            lappend path [$W get]
        } else {
            # 0 1 2
            #   ^
            # 0 1
            #     2
            for {set i $no; incr i} {$i < $l} {incr i} {
                pack forget $pane.presets.add.cb$i
            }
            set path [lreplace $path $no end [$W get]]
        }
        my makecombo $path
    }

    method updatetags {tags {update rewrite}} {
        global modulesd
        [dict get $modulesd TagEdit] updatetags $tags {*}$update
        pack forget {*}[info commands $pane.presets.add.cb*]
        pack $pane.presets.add.cb0 -fill x -before $pane.presets.add.details
        return
        if {$update eq "update"} {
            set item [$pane.tags focus]
            if {$item ne ""} {
                $pane.tags item $item -values $tags
            }
        } else {
            $pane.tags delete [$pane.tags children {}]
            foreach {key value} $tags {
                $pane.tags insert {} end -values [list $key $value]
            }
            $pane.tags insert {} end -values {{} {}}
        }
    }

    method edittag {x y} {
        puts [set values [$pane.tags item [$pane.tags identify row $x $y] -values]]
        my tagpopup .tagedit $values
    }

    method updatetag values {
        puts $values
    }

    method tagpopup {w values} {
        lassign $values tagk tagv
        toplevel $w
        ttk::label $w.l1 -text Key
        ttk::entry $w.e1 -textvariable [my varname tagk]
        ttk::label $w.l2 -text Value
        ttk::entry $w.e2 -textvariable [my varname tagv]
        grid $w.l1 -row 0 -column 0 -sticky w -padx 4 -pady 4
        grid $w.e1 -row 0 -column 1 -sticky we -padx 4 -pady 4
        grid $w.l2 -row 1 -column 0 -sticky w -padx 4 -pady 4
        grid $w.e2 -row 1 -column 1 -sticky we -padx 4 -pady 4
        grid rowconfigure $w $w.e1 -weight 1
        grid rowconfigure $w $w.e2 -weight 1
        grid columnconfigure $w $w.e1 -weight 1
        grid columnconfigure $w $w.e2 -weight 1
        ttk::frame $w.buttons
        ttk::button $w.buttons.ok -text OK -command [string map [list %W $w %my [namespace which my] %k [my varname tagk] %v [my varname tagv]] {
            wm withdraw %W
            rename %W {}
            %my updatetags [list $%k $%v] update
        }] -default active
        ttk::button $w.buttons.cancel -text Cancel -command [string map [list %W $w %values $values] {
            wm withdraw %W
            rename %W {}
            # %my updatetags [list %values]
        }]
        pack $w.buttons.cancel $w.buttons.ok -side left -padx 4
        grid $w.buttons - -row 2
        wm resizable $w 0 0
        wm withdraw $w
        update
        set x [expr {([winfo screenwidth .]-[winfo width $w])/2}]
        set y [expr {([winfo screenheight .]-[winfo height $w])/2}]
        wm geometry  $w +$x+$y
        wm transient $w .
        wm title     $w "Change tags"
        wm deiconify $w
        wm resizable $w 1 1
    }
}

