#!/usr/bin/env tclsh

coroutine readline apply {{s} {
    yield
    set l [split $s \n]
    foreach line $l {
        yield [list DATA $line]
    }
    yield END
}} [read [open test.mapcss r]]

coroutine gettoken apply {{} {
    set s {}
    set r {}
    set end 0
    set state BEGIN
    while 1 {
        set y [yield $r]
        # puts "y=$y"
        if {$y ne "END"} {
            set s "$s[lindex $y 1] "
        } else {
            set end 1
        }
        # puts "s=$s, state=$state"
        # puts "state=$state"
        switch $state {
            BEGIN {
                set s [string trimleft $s]
                if {([string length $s] == 0) && ($end)} {
                    yield END
                }
                # puts "trimmed=$s"
                switch -glob $s {
                    {/\**} {
                        # puts comment?
                        set cs 0
                        set ce [string first {*/} $s $cs+2]
                        set r ...COMMENT...
                        while {($ce == -1) && (!$end)} {
                            set y [yield $r]
                            if {$y ne "END"} {
                                set s "$s[lindex $y 1] "
                            } else {
                                set end 1
                            }
                            set ce [string first {*/} $s $cs+2]
                        }
                        set s [string replace $s $cs $ce+1 {}]
                        set r COMMENT
                    }
                    {canvas*} -
                    {way*} -
                    {node*} -
                    {relation*} {
                        if {[regexp {^(\w+)(?:\|z(\d+)?(-)?(\d+)?)?} $s match type z1 zdash z2]} {
                            set s [string replace $s 0 [string length $match]-1 {}]
                            set state SELECTOR
                            if {$zdash ne "-"} {
                                if {$z1 eq ""} {
                                    set z1 $z2
                                }
                                if {$z2 eq ""} {
                                    set z2 $z1
                                }
                                if {$z2 eq ""} {
                                    set z1 1
                                    set z2 end
                                }
                            } else {
                                if {$z1 eq ""} {
                                    set z1 1
                                }
                                if {$z2 eq ""} {
                                    set z2 end
                                }
                            }
                            # puts "object: $type $z1 $z2"
                            set r [list OBJECT $type $z1 $z2]
                        }
                    }
                    {.*} {
                        if {[regexp {.(\w+)} $s match inv pseudo]} {
                            set s [string replace $s 0 [string length $match]-1 {}]
                            set r [list CLASS $inv$pseudo]
                            set state SELECTOR
                        }
                    }
                    {:*} -
                    {!:*} {
                        if {[regexp {(!)?:(\w+)} $s match inv pseudo]} {
                            set s [string replace $s 0 [string length $match]-1 {}]
                            set r [list PSEUDO $inv$pseudo]
                        }
                    }
                    {\{*} {
                        set s [string trimleft [string replace $s 0 0 {}]]
                        set state DECLARATION
                        set r DECLARATION...
                    }
                }
            }
            SELECTOR {
                # puts @@$s@@
                if {[string match {[ \r\n\t]*} $s]} {
                    set state BEGIN
                    set r ENDSELECTOR
                } else {
                    set s [string trimleft $s]
                    switch -glob $s {
                        {,*} {
                            set r {AND}
                            set s [string replace $s 0 0 {}]
                            set state BEGIN
                        }
                        {\[*} {
                            if {[regexp {^\[(?:(!)?([a-z:_0-9]+)[ \t\r\n]*(?:(!=|=|<=|>=|>|<)[ \t\r\n]*(?:([a-zA-Z0-9:_*,./-]+)|"([^\x22]*)"))?)\]} $s \
                                match inv tag op val qval]} {
                                # "
                                if {$op eq ""} {
                                    set op =
                                    set val yes
                                }
                                if {$val eq ""} {
                                    set val $qval
                                }
                                if {$inv eq "!"} {
                                    switch $op {
                                        "<" {
                                            set op >=
                                        }
                                        ">" {
                                            set op <=
                                        }
                                        "=" {
                                            set op !=
                                        }
                                        "!=" {
                                            set op =
                                        }
                                        "<=" {
                                            set op >
                                        }
                                        ">=" {
                                            set op <
                                        }
                                    }
                                }
                                set s [string replace $s 0 [string length $match]-1 {}]
                                # puts ">$type $z1 $z2 $tag $inv $op $val $qval"
                                set r [list SELECTOR $tag $op $val]
                            }
                        }
                        {:*} {
                            if {[regexp {(!)?:(\w+)} $s match inv pseudo]} {
                                set s [string replace $s 0 [string length $match]-1 {}]
                                set r [list PSEUDO $inv$pseudo]
                            }
                        }
                        {\{*} {
                            set s [string trimleft [string replace $s 0 0 {}]]
                            set state DECLARATION
                            set r DECLARATION...
                        }
                    }
                }
            }
            DECLARATION {
                set s [string trimleft $s]
                switch -glob $s {
                    {/\**} {
                        # puts comment?
                        set cs 0
                        set ce [string first {*/} $s $cs+2]
                        set r ...COMMENT...
                        while {$ce == -1} {
                            set s $s[yield $r]
                            set ce [string first {*/} $s $cs+2]
                        }
                        set s [string replace $s $cs $ce+1 {}]
                        set r COMMENT
                    }
                    {;*} {
                        set s [string replace $s 0 0 {}]
                        set r ...DECLARATION...
                    }
                    {\}*} {
                        set s [string replace $s 0 0 {}]
                        set r ...DECLARATION
                        set state BEGIN
                    }
                    {exit} {
                        set r [list DECLARATION exit]
                        set ce [string first ; $s]
                        set s [string replace $s 0 $ce {}]
                    }
                    {set} {
                        set ce [string first ; $s]
                        set r [list DECLARATION set [string range $s 4 $ce]]
                        set s [string replace $s 0 $ce {}]
                    }
                    default {
                        if {[regexp {^(([a-z-]+)[ ]*:[ ]*(?:(?:"([^\x22]*)"|([0-9,. ]+)|(#[,0-9a-fA-F]+)+|([A-Za-z_0-9/.-]+))[ \t\r\n]*[;\x7d]))} $s -> match key qval val hval idval]} {
                            if {$idval ne ""} {
                                set val $idval
                            }
                            if {$hval ne ""} {
                                set val [string tolower $hval]
                            }
                            if {$qval ne ""} {
                                set val $qval
                            }
                            set r [list DECLARATION $key $val]
                            set s [string replace $s 0 [string length [string trimright $match ";\x7d"]]-1 {}]
                        } elseif {[regexp {^([a-z-]+)[ ]*:[^;\x7d]+} $s match key]} {
                            set r [list DECLARATION $key EVAL]
                            set s [string replace $s 0 [string length $match]-1 {}]
                        }
                    }
                }
            }
        }
    }
}}

while 1 {
    if {[catch {readline} rl]} {
        set rl END
    }
    # puts "get=$r"
    set r [gettoken $rl]
    if {$r eq "END"} {
        return
    } else {
        # puts "R: $r"
        lassign $r -> t
        if {![string match {*...*} $r]} {
            puts "R: $r"
        }
    }
}

# vim: set ts=4:sw=4:sts=4
